// Copyright 2023 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package jobqueue

import (
	"errors"
	"gorm.io/gorm"
	"sync"
)

type DBSaverStatus int

const (
	DBSaverStatusStopped = iota
	DBSaverStatusRunning
)

type DBSaver struct {
	saveChannel chan GenericJob
	stopChan    chan struct{}
	migrateFlag bool
	manager     *Manager
	status      DBSaverStatus
	mu          sync.Mutex
}

type RunnerData string
type SchedulerData string

// NewDBSaver creates a new DBSaver
func NewDBSaver() *DBSaver {
	return &DBSaver{
		saveChannel: make(chan GenericJob, 100),
		stopChan:    make(chan struct{}),
	}
}

// SetManager sets the manager of the DBSaver
func (s *DBSaver) SetManager(manager *Manager) *DBSaver {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.manager = manager
	return s
}

func (s *DBSaver) setStatus(status DBSaverStatus) *DBSaver {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.status = status
	return s
}

// isStatus returns true if the DBSaver has the given status
// the lock is not needed here, because it is only used in the Start() method
func (s *DBSaver) isStatus(status DBSaverStatus, lock bool) bool {
	if lock {
		s.mu.Lock()
		defer s.mu.Unlock()
	}

	return s.status == status

}

// Start starts the DBSaver
func (s *DBSaver) Start() error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	if s.isStatus(DBSaverStatusRunning, false) {
		return nil
	}

	db := s.manager.database

	if !s.migrateFlag {
		err := db.AutoMigrate(&JobPersistence{}, &JobLog{}, &JobStats{})
		if err != nil {
			return err
		}
		s.migrateFlag = true
	}

	var wg sync.WaitGroup
	wg.Add(1)

	go func() {
		wg.Done()

		// this is protected by the lock above
		s.status = DBSaverStatusRunning

		defer func() {
			// this runs after the function returns
			// and needs to be protected by the lock
			// of the setStatus method
			s.status = DBSaverStatusStopped
			//s.setStatus(DBSaverStatusStopped)
		}()

		for {
			select {
			case job := <-s.saveChannel:

				err := db.Transaction(func(tx *gorm.DB) error {

					permJob := job.GetPersistence()

					memLogs := permJob.Logs
					permJob.Logs = nil

					var existingJob JobPersistence
					result := tx.Unscoped().Where("id = ?", permJob.GetID()).First(&existingJob)

					if result.Error != nil {
						if errors.Is(result.Error, gorm.ErrRecordNotFound) {
							err := tx.Create(&permJob).Error
							if err != nil {
								return err
							}
						} else {
							return result.Error
						}
					} else {

						db.Unscoped().Model(&existingJob).Update("deleted_at", nil)

						tx.Model(&existingJob.Scheduler).Select(
							[]string{
								"type",
								"interval",
								"spec",
								"delay",
								"event",
								"time",
								"executed",
							}).UpdateColumns(SchedulerPersistence{
							Type:     "",
							Interval: 0,
							Spec:     "",
							Delay:    0,
							Event:    "",
							Time:     nil,
							Executed: false,
						})

						err := tx.Model(&existingJob).Updates(permJob).Error
						if err != nil {
							return err
						}
					}

					tx.Model(&permJob.Stats).
						Select(
							[]string{
								"run_count",
								"success_count",
								"error_count",
								"time_metrics_avg_run_time",
								"time_metrics_max_run_time",
								"time_metrics_min_run_time",
								"time_metrics_total_run_time",
							},
						).
						UpdateColumns(permJob.Stats)

					for i, _ := range memLogs {
						memLogs[i].LogID = 0
						_ = tx.Create(&memLogs[i])
						// no error handling, if it fails, it fails
					}

					return nil // Commit the transaction
				})

				if err != nil {
					s.logError("Error while saving job", "error", err)
				}

			case <-s.stopChan:
				return
			}
		}
	}()

	wg.Wait()

	return nil
}

// logError logs an error
func (s *DBSaver) logError(msg string, keysAndValues ...interface{}) {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.logger == nil {
		return
	}

	s.manager.logger.Error(msg, keysAndValues...)
}

// Stop stops the DBSaver
func (s *DBSaver) Stop() *DBSaver {
	s.mu.Lock()
	defer s.mu.Unlock()

	select {
	case s.stopChan <- struct{}{}:
	default:
		s.logError("DBSaver stop channel is full")
	}

	return s
}

// SaveJob saves a job to the database
func (s *DBSaver) SaveJob(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.saveChannel == nil {
		return ErrDBSaverNotInitialized
	}

	if s.status != DBSaverStatusRunning {
		return ErrDBSaverNotRunning
	}

	defer func() {
		if r := recover(); r != nil {
			s.logError("Error while saving job", "error", r)
		}
	}()

	select {
	case s.saveChannel <- job:
	default:
		// if the channel is full, we just drop the job
		// this is not ideal, but better than blocking
		// the job queue
		s.logError("DBSaver channel is full, dropping job with  ID", "job_id", job.GetID())
	}

	return nil
}

func checkRunningSaver(s *DBSaver) (*gorm.DB, error) {
	if s.manager == nil {
		return nil, ErrNoManager
	}

	if s.manager.database == nil {
		return nil, ErrNoDatabaseConnection
	}

	if !s.isStatus(DBSaverStatusRunning, false) {
		return nil, ErrDBSaverNotRunning
	}

	return s.manager.database, nil

}

// DeleteJob deletes a job from the database
func (s *DBSaver) DeleteJob(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()
	var db *gorm.DB
	var err error

	if db, err = checkRunningSaver(s); err != nil {
		return err
	}

	return db.Transaction(func(tx *gorm.DB) error {
		permJob := job.GetPersistence()

		dbErr := tx.Where("job_id = ?", permJob.GetID()).Delete(&JobLog{}).Error
		if dbErr != nil {
			return dbErr
		}

		dbErr = tx.Where("job_id = ?", permJob.GetID()).Delete(&JobStats{}).Error
		if dbErr != nil {
			return dbErr
		}

		dbErr = tx.Delete(&permJob).Error
		if dbErr != nil {
			return dbErr
		}

		return nil
	})

}

func (s *DBSaver) ResetLogs(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()
	var db *gorm.DB
	var err error

	if db, err = checkRunningSaver(s); err != nil {
		return err
	}

	return db.Transaction(func(tx *gorm.DB) error {
		permJob := job.GetPersistence()

		// unscoped because we want to delete the logs finally
		dbErr := tx.Unscoped().Where("job_id = ?", permJob.GetID()).Delete(&JobLog{}).Error
		if dbErr != nil {
			return dbErr
		}

		return nil
	})
}

func (s *DBSaver) ResetStats(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.saveChannel == nil {
		return ErrDBSaverNotInitialized
	}

	if s.status != DBSaverStatusRunning {
		return ErrDBSaverNotRunning
	}

	defer func() {
		if r := recover(); r != nil {
			s.logError("Error while saving job", "error", r)
		}
	}()

	job.ResetStats()

	select {
	case s.saveChannel <- job:
	default:
		s.logError("DBSaver channel is full, dropping job with  ID", "job_id", job.GetID())
	}

	return nil
}
